
; Code to turn PROC lambdas into JUMP lambdas.

(define-subrecord lambda jump-block lambda-block
  ()
  (
   parent
   children
   calls
   status
  ))

(define make-jump-block jump-block-maker)

; 1) Determine the the parent-children relationships of the lambdas using the
; LAMBDA-BLOCK fields.
; 2) Call REALLY-INTEGREATE-JUMP-PROCS on each proc.
; 3) Clear the LAMBDA-BLOCK fields.
; 4) Return #t if anything changed.

(define (integrate-jump-procs)
  (let ((lambdas (make-lambda-list)))
    (walk (lambda (l)
	    (set-lambda-block! l (make-jump-block)))
	  lambdas)
    (receive (procs others)
	(find-scoping lambdas
		      lambda-parent set-lambda-parent!
		      lambda-children set-lambda-children!)
      (let ((change? (or-map really-integrate-jump-procs procs)))
	(walk (lambda (l)
		(set-lambda-block! l #f))
	      lambdas)
	change?))))
  
; Try each child in turn to see if it can be changed to a JUMP procedure.
; This may involve making several procedures JUMP procedures if they are
; mutually recursive.
; A PROC lambda may be turned into a JUMP lambda if all its calling points are
; known and call only that lambda, and all its calling points are within it
; enclosing procedure or within other procedures that can be turned into JUMP
; lambdas.

(define (really-integrate-jump-procs proc)
  (let ((procs (filter (lambda (l)
			 (and (eq? 'proc (lambda-type l))
			      (calls-known? l)))
		       (lambda-children proc))))
    (walk (lambda (p)
	    (set-lambda-calls! p (find-calls p))
	    (set-lambda-status! p 'unseen))
	  procs)
    (let loop ((procs procs) (c? #f))
      (if (null? procs)
          c?
	  (let ((p (car procs)))
	    (receive (jump-procs cont)
		(if (eq? (lambda-status p) 'unseen)
		    (find-jump-group p proc)
		    (values #f #f))
	      (cond (jump-procs
		     (walk (lambda (p) (set-lambda-status! p 'integrated))
			   jump-procs)
		     (actually-integrate-jump-procs jump-procs proc cont)
		     (loop (cdr procs) #t))
		    (else
		     (set-lambda-status! p #f)
		     (loop (cdr procs) c?)))))))))

; Checks to see if PROC is a member of a group of procedures called with only
; one non-trivial continuation.  The continuation must be from within the
; PROC's parent and all the other procs must be children of the same parent.
; REPS is a list of the representations of the values passed to CONT.

(define (find-jump-group proc parent)
  (let loop ((to-do (list proc)) (done '()) (parent-calls '()) (reps 'none))
    (cond ((null? to-do)
	   (check-calls-in-parent done parent-calls reps))
	  (else
	   (let ((proc (car to-do)))
	     (set-lambda-status! proc 'checked)
	     (receive (ok? new-to-do parent-calls reps)
		 (check-proc-calls proc (cdr to-do) parent-calls reps parent)
	       (cond (ok?
		      (loop new-to-do (cons proc done) parent-calls reps))
		     (else
		      (walk mark-lambda-no-good to-do)
		      (walk mark-lambda-no-good done)
		      (values #f #f)))))))))

(define (mark-lambda-no-good proc)
  (set-lambda-status! proc 'no-good))

(define (check-proc-calls proc to-do parent-calls reps parent)
  (let loop ((refs (lambda-calls proc))
	     (to-do to-do) (parent-calls parent-calls) (reps reps))
    (cond ((null? refs)
	   (values #t to-do parent-calls reps))
	  ((eq? parent (proc-ancestor (car refs)))
	   (loop (cdr refs) to-do
		 (cons (node-parent (car refs)) parent-calls)
		 reps))
	  (else
	   (receive (ok? proc reps)
	       (check-tail-call (car refs) parent reps)
	     (if ok?
		 (loop (cdr refs)
		       (if proc (cons proc to-do) to-do)
		       parent-calls
		       reps)
		 (values #f #f #f #f)))))))

(define (check-tail-call ref parent reps)
  (let* ((proc (proc-ancestor ref))
	 (status (lambda-status proc)))
    (if (not (or (eq? status 'unseen) ; anything but #f, INTEGRATED, or  NO-GOOD
		 (eq? status 'checked)
		 (eq? status 'seen)))
	(values #f #f #f)
	(receive (ok? reps)
	    (call-is-tail-recursive ref reps)
	  (values ok?
		  (cond ((eq? status 'unseen)
			 (set-lambda-status! proc 'seen)
			 proc)
			(else #f))
		  reps)))))

(define (call-is-tail-recursive ref reps)
  (let* ((cont (call-arg (node-parent ref) 0))
	 (next-call (lambda-body cont))
	 (vars (lambda-variables cont)))
    (cond ((or (neq? 'return (primop-id (call-primop next-call)))
	       (not (return-uses-these-variables? vars next-call)))
	   (values #f #f))
	  ((eq? reps 'none)
	   (values #t (map variable-type vars)))
	  ((vars-use-these-reps? vars reps)
	   (values #t reps)))))

(define (return-uses-these-variables? vars return)
  (let loop ((vars vars) (i 1))
    (cond ((>= i (call-arg-count return))
	   (null? vars))
	  ((or (null? vars)
	       (not (reference-node? (call-arg return i)))
	       (neq? (car vars)
		     (reference-variable (call-arg return i))))
	   #f)
	  (else
	   (loop (cdr vars) (+ i 1))))))

(define (vars-use-these-reps? vars reps)
  (let loop ((vars vars) (reps reps))
    (cond ((null? vars)
	   (null? reps))
	  ((or (null? reps)
	       (not (type-eq? (variable-type (car vars)) (car reps))))
	   #f)
	  (else
	   (loop (cdr vars) (cdr reps))))))

(define (check-calls-in-parent procs calls reps)
  (let loop ((calls calls) (cont #f))
    (cond ((null? calls)
	   (if (and cont
		    (or (eq? reps 'none)
			(vars-use-these-reps? (lambda-variables cont) reps)))
	       (values procs cont)
	       (values #f #f)))       ; this is an odd possiblity
	  ((not cont)
	   (loop (cdr calls) (call-arg (car calls) 0)))
	  ((node-equal? (call-arg (car calls) 0) cont)
	   (loop (cdr calls) cont))
	  (else
	   (values #f #f)))))
	   
;=============================================================================;
; Part 2.  Turn JUMP-PROCS from procs to jumps.  All are children of PROC and
; CONT is the actual continuation, which is also turned into a jump.

; This finds the least-common-ancestor of all of the calls to the
; jump-procs and binds them and the continuation with a LETREC at that
; point.

(define (actually-integrate-jump-procs jump-procs proc cont)
  (let* ((vars (map bound-to-variable jump-procs))
	 (lca (least-common-ancestor (apply append (map variable-refs vars))))
	 (cvar (make-variable 'w (node-type cont))))
;    (if (not (eq? 'letrec2 (primop-id (call-primop lca))))
;	(breakpoint "funny LCA for jump procs"))
    (for-each (lambda (var)
		(let ((binder (variable-binder var)))
		  (set-lambda-variables! binder (delq! var (lambda-variables binder)))))
	      vars)
    (for-each (lambda (node)
		(let ((parent (node-parent node))
		      (index (node-index node)))
		  (detach node)
		  (remove-call-arg parent index)))
	      jump-procs)
    (move cont (lambda (ignore) (make-literal-node '#f '#f)))
    (put-in-letrec (cons cvar vars) (cons cont jump-procs) lca)
    (for-each proc-calls->jumps jump-procs)
    (change-lambda-type cont 'jump)
    (walk (lambda (p)
	    (let* ((v (car (lambda-variables p)))
		   (refs (variable-refs v)))
	      (set-variable-refs! v '())
	      (walk (lambda (r)
		      (return->jump (node-parent r) cvar))
		    refs)
	      (remove-variable p v)))
	  jump-procs)
    (values)))

(define (put-in-letrec vars values call)
  (cond ((eq? 'letrec2 (primop-id (call-primop call)))
	 (let ((binder (node-parent call)))
	   (mark-changed call)
	   (for-each (lambda (var)
		       (set-variable-binder! var binder))
		     vars)
	   (set-lambda-variables! binder
				  (append (lambda-variables binder) vars))
	   (for-each (lambda (value)
		       (append-call-arg call value))
		     values)))
	(else
	 (move-body
	  call
	  (lambda (call)
	    (let-nodes ((c (letrec1 1 l2))
			(l2 ((x #f) . vars) (letrec2 1 l3 (* x) . values))
			(l3 () call))
	      c))))))

; Put CONT into a LETREC at the beginning of TOP-NODE's body.
; (lambda (c ...) ... <cont> ...)
;  =>
; (lambda (c ...)
;   (let ((w (letrec))
;     ...
;     ($set-contents w <cont>)
;     ...))

;(define (put-cont-in-letrec top-node cont)
;  (let ((var (make-variable 'w (node-type cont))))
;    (move-body
;     (lambda-body top-node)
;     (lambda (call)
;       (let-nodes ((c (letrec 1 l1)) 
;                   (l1 ('var) call))
;         c)))
;    (move-body
;     (node-parent cont)
;     (lambda (call)
;       (move cont (lambda (ignore) (make-literal-node '#f '#f)))
;       (let-nodes ((c (set-contents 1 l1 (* var) '('letrec #f)
;				    '((node-type cont) #f) cont))
;                   (l1 () call))
;         c)))
;    var))

; Turn all calls to PROC into jumps.

(define (proc-calls->jumps proc)
  (walk (lambda (n)
          (call->jump (node-parent n)))
        (lambda-calls proc))
  (change-lambda-type proc 'jump))

; Change a call to a jump by changing the primop and removing the continuation.

(define (call->jump call)
  (set-call-primop! call (get-primop primop/jump))
  (remove-call-arg call 0))

; Change a return to a jump.  VAR is a variable whose value is a cell containing
; the procedure to which to jump.

(define (return->jump call var)
  (set-call-primop! call (get-primop primop/jump))
  (replace (call-arg call 0) (make-reference-node var)))
