
; Evaluation of top-level forms.

; Entry point
(define (evaluate exp)
  (eval-exp exp (empty-env)))

(define evaluator-table (make-table))

(define (eval-exp exp env)
  (cond ((variable? exp)
	 (env-ref env exp))
	((quote-exp? exp)
	 (quote-exp-value exp))
	((let-exp? exp)
	 (eval-exp (let-exp-body exp)
		   (env-bind* env
			      (let-exp-vars exp)
			      (evaluate-list (let-exp-vals exp) env))))
	((call-exp? exp)
	 (eval-primop-call (call-exp-proc exp)
			   (call-exp-exits exp)
			   (call-exp-type exp)
			   (call-exp-args exp)
			   env))
	((block-exp? exp)
	 (do ((exps (block-exp-exps exp) (cdr exps))
	      (value the-undefined-value (eval-exp (car exps) env)))
	     ((null? exps)
	      value)))
	((lambda-exp? exp)
	 (make-closure exp env))
	((return-exp? exp)
	 (eval-exp (car (return-exp-args exp)) env))
	((letrec-exp? exp)
	 (let* ((vars (letrec-exp-vars exp))
		(env (env-bind* env
				vars
				(map (lambda (x) the-undefined-value) vars)))
		(vals (evaluate-list (letrec-exp-vals exp) env)))
	   (for-each (lambda (var val)
		       (env-set! env var val))
		     vars
		     vals)
	   (eval-exp (letrec-exp-body exp) env)))
        (else
         (bug "unknown syntax~% ~S" exp))))

(define (evaluate-list exps env)
  (map (lambda (exp)
	 (eval-exp exp env))
       exps))

(define (apply-lambda-exp exp args env)
  (eval-exp (lambda-exp-body exp)
	    (env-bind* env
		       (lambda-exp-vars exp)
		       args)))

(define (apply-closure closure args)
  (if (not (closure? closure))
      (error "attempt to call a non-procedure ~S" closure)
      (let ((exp (closure-exp closure)))
	(if (not (= (length args) (length (lambda-exp-vars exp))))
	    (error "wrong number of arguments in call ~S" (cons closure args))
	    (apply-lambda-exp exp args (closure-env closure))))))

; Primops that return one value (many of them) use this to handle the
; absence or presence of an explicit continuation.

(define (eval-call exits args env proc)
  (case exits
    ((0)
     (proc args))
    ((1)
     (apply-lambda-exp (car args) (list (proc (cdr args))) env))
    (else
     (bug "bad number of exits ~S in call" exits))))

(define (eval-call! exits args env proc)
  (case exits
    ((0)
     (proc args)
     the-undefined-value)
    ((1)
     (proc (cdr args))
     (apply-lambda-exp (car args) '() env))
    (else
     (bug "bad number of exits ~S in call" exits))))

;--------------------------------------------------
; Environments are a-lists.

(define (env-ref env var)
  (let ((value (cond ((assq var env)
		      => cdr)
		     ((variable-form var)
		      => form-value)
		     (else
		      (error "unbound variable ~S" var)))))
    (if (undefined-value? value)
	(error "reference to uninitialized variable ~S" var)
	value)))

(define (env-set! env var value)
  (cond ((assq var env)
	 => (lambda (p)
	      (set-cdr! p value)))
	((variable-form var)
	 => (lambda (form)
	      (set-form-value! form value)))
	(else
	 (error "unbound variable ~S" var))))

(define (empty-env)
  '())

(define (env-bind env var val)
  (cons (cons var val) env))

(define (env-bind* env vars vals)
  (append (map cons vars vals) env))

;--------------------------------------------------

(define (stored-object-ref object type)
  (let ((object (stob-value object)))
    (cond ((vector? object)
	   (vector-ref object 0))
	  ((eq? type type/int8)
	   (byte-vector-ref object 0))
	  ;((eq? type type/int16)
	  ; (byte-vector-half-word-ref object offset))
	  ((eq? type type/int32)
	   (byte-vector-word-ref object 0))
	  (else
	   (bug "bad type in referencing bytes ~S" type)))))
		      
(define (stored-object-set! object type value)
  (let ((object (stob-value object)))
    (cond ((vector? object)
	   (vector-set! object 0 value))
	  ((eq? type type/int8)
	   (byte-vector-set! object 0 value))
	  ;((eq? type type/int16)
	  ; (byte-vector-half-word-set! object offset value))
	  ((eq? type type/int32)
	   (byte-vector-word-set! object 0 value))
	  (else
	   (bug "bad type in setting bytes ~S" type)))))
		      
(define the-uninitialized-value (list '*uninitialized-value*))

(define (uninitialized-value? x)
  (eq? x the-uninitialized-value))

