
; External entry point

(define (expand-node node package)
  (set! *free-variables* '())
  (receive (id var exp)
      (if (define-node? node)
	  (let ((exp (node-form node)))
	    (values (cadr exp)
		    (get-package-variable package (cadr exp))
		    (caddr exp)))
	  (values #f #f node))
    (let ((cenv ((structure-ref packages package->environment) package)))
      (prepare-cenv package cenv)
      (let* ((exp (expand exp cenv id))
	     (vars *free-variables*))
	(set! *free-variables* '())
	(for-each (lambda (v) (set-variable-flag! v #f)) vars)
	(values var exp vars)))))

(define define-node? (node-predicate 'define syntax-type))

(define (get-package-variable package id)
  (let ((value ((structure-ref packages package-lookup) package id)))
    (if (and (binding? value)
	     (variable? (binding-place value)))
	(binding-place value)
	;; this is what will happen if + is used in a non-call position
	(error "defined name ~S not bound to a variable ~S"
	       id (binding-place value)))))

; Main dispatch for expanding a single expression (and internal entry point)

(define (expand exp cenv id)
  (really-expand-node (classify exp cenv) cenv id))

(define (really-expand-node node cenv id)
  ((operator-table-ref compilators (node-operator-id node))
   node
   cenv
   id))

(define compilators
  (make-operator-table (lambda (node cenv id)
			 (bug "not valid in expression context ~S"
			      (schemify node)))
		       (lambda (frob)  ;for let-syntax, with-aliases, etc.
			 (lambda (node cenv id)
			   (call-with-values (lambda () (frob node cenv))
			     (lambda (form cenv)
			       (expand form cenv id)))))))

(define (define-compilator name proc)
  (operator-define! compilators name proc))

;(define (expand-invalid-context form cenv id)
;  (syntax-error "form not appropriate in expression context" form))

(define-local-syntax (define-scheme-syntax pattern . stuff)
  (receive (name destructure-clause vars body)
      (if (pair? pattern)
	  (destructure (((vars . body) stuff))
	    (values (car pattern)
		    `(,(cdr pattern) (cdr (node-form xnode)))
		    vars
		    body))
	  (destructure (((node-var vars . body) stuff))
	    (values pattern `(,node-var xnode) vars body)))
    `(operator-define! compilators ',(case name  ; hack around bizarre interface
				       ((name literal call unspecific) name)
				       (else (list name syntax-type)))
		       (lambda (xnode . ,vars)
			 (destructure (,destructure-clause)
			   ,@body)))))

; QUOTE

(define-scheme-syntax (quote exp) (cenv id)
  (expand-constant exp))

(define-scheme-syntax literal node (cenv id)
  (expand-constant (node-form node)))

(define (expand-constant value)
  (let ((value (cond ((integer? value) value)
		     ((char?    value) value)
		     ((eq? value #t)   true-value)
		     ((eq? value #f)   false-value)
		     (else             value))))
    (make-quote-exp value type/unknown)))

(define-scheme-syntax unspecific some-node (cenv id)
  (make-quote-exp (structure-ref node the-undefined-value) type/unknown))

; Variables

(define-scheme-syntax name node (cenv id)
  (let ((var (get-variable (node-form node)
			   (name-node-binding node)
			   cenv)))
    (cond ((global-variable? var)
	   (note-global-use! var)
	   (if (variable-settable? var)
	       (dereference-cell #f var (variable-type var) 'global)
	       var))
	  (else
	   (if (variable-flag var)
	       (note-letrec-var-use var))
	   var))))

(define (get-variable name binding cenv)
  (cond ((variable? binding)   ; local variables
	 binding)
	((binding? binding)
	 (let ((place (binding-place binding)))
	   (cond ((variable? place)
		  place)
		 ((primitive? place)
		  (primitive-variable (binding-place binding)))
		 (else
		  (error "unbound variable ~S in ~S" name (cenv-package cenv))))))
	(else
	 (error "unbound variable ~S in ~S" name (cenv-package cenv)))))

; Stolen from bcomp/ctop

(define (prepare-cenv package env)
  (bind1 *the-package* package env))
(define (cenv-package env)
  (lookup env *the-package*))
(define *the-package* (string->symbol "The package"))

(define *free-variables* '())

(define (note-global-use! var)
  (cond ((not (variable-flag var))
	 (set! *free-variables* (cons var *free-variables*))
	 (set-variable-flag! var #t))))

; SET!

(define-scheme-syntax set! node (cenv id)
  (destructure (((#f lhs rhs) (node-form node)))
    (let* ((lhs-node (classify lhs cenv))
	   (var (get-variable (node-form lhs-node)
			      (name-node-binding lhs-node)
			      cenv)))
      (if (variable-settable? var)
	  (let ((value (expand rhs cenv (variable-name var))))
	    (make-block-exp (list (set-cell (node-form node)
					    var value type/unknown 'global)
				  (make-undefined-value))))
	  (error "illegal variable set ~S" `(set! ,lhs ,rhs))))))

(define (variable-settable? var)
  (let ((name (variable-name var)))
    (and (global-variable? var)
	 (let ((name (symbol->string name)))
	   (and (eq? '#\* (string-ref name 0))
		(eq? '#\* (string-ref name (- (string-length name) 1)))
		(<= 3 (string-length name)))))))

; LAMBDA

(define-scheme-syntax lambda node (cenv id)
  (destructure (((#f formals . body) (node-form node)))
    (if (n-ary? formals)
	(error "n-ary procedures are not allowed in PreScheme (LAMBDA ~S ...)"
	       formals))
    (let* ((vars (map (lambda (name)
			(if (not name)
			    #f
			    (make-variable (name->symbol name) type/unknown)))
		      formals))
	   (value (expand-block body (bind formals vars cenv) #f)))
      (make-lambda-exp id
		       #f
		       (length vars)
		       vars
		       (make-return-exp 1 type/unknown (list value))
		       (node-form node)))))

; IF

(define-scheme-syntax if node (cenv id)
  (let ((exp (node-form node)))
    (destructure (((test consequence alternate) (cdr exp)))
      (let ((test (expand test        cenv #f))
	    (cons (expand consequence cenv id))
	    (alt  (expand alternate   cenv id)))
	(make-call-exp (lookup-primop 'test)
		       2
		       type/unknown
		       (list (make-continuation-exp '() cons)
			     (make-continuation-exp '() alt)
			     test)
		       exp)))))

; BEGIN

(define-scheme-syntax (begin . forms) (cenv id)
  (expand-block forms cenv id))

(define (expand-block forms cenv id)
  (scan-body forms
	     cenv
	     (lambda (defs exps)
	       (if (null? defs)
		   (really-expand-block exps cenv id)
		   (expand-letrec (map (lambda (node)
					 (cdr (node-form node)))
				       defs)
				  exps
				  cenv
				  id)))))

(define (really-expand-block forms cenv id)
  (make-block-exp
   (if (null? forms)
       (error "empty block")
       (do ((fs forms (cdr fs))
	    (exps '() (cons (expand (car fs) cenv #f) exps)))
	   ((null? (cdr fs))
	    (reverse (cons (expand (car fs) cenv id)
			   exps)))))))

; Applications

(define-scheme-syntax call node (cenv id)
  (let* ((exp (node-form node))
	 (proc (classify (car exp) cenv))
	 (args (cdr exp)))
    (expand-call exp proc args cenv id #f)))

(define-scheme-syntax goto node (cenv id)
  (let* ((exp (node-form node))
	 (proc (classify (cadr exp) cenv))
	 (args (cddr exp)))
    (expand-call exp proc args cenv id #t)))

(define (expand-call exp proc args cenv id goto?)
    (cond ((lambda-node? proc)
	   (expand-redex exp proc (cdr exp) cenv id))
	  ((name-node? proc)
	   (expand-call-to-name proc args exp goto? cenv))
	  (else
	   (expand-unknown-call exp proc args goto? cenv))))

(define (expand-call-to-name proc args exp goto? cenv)
  (let ((binding (name-node-binding proc)))
    (cond ((variable? binding)  ; local variable
	   (expand-unknown-call exp proc args goto? cenv))
	  ((generated? binding)
	   (let ((var ((structure-ref environments environment-ref)
		       (generated-env binding)
		       (generated-symbol binding))))
	     (expand-unknown-call var proc args goto? cenv)))
	  ((symbol? binding)
	   (error "unbound variable ~S" binding))
	  ((not (binding? binding))
	   (error "unknown type of binding ~S" binding))
	  (else
	   (let ((value (binding-place binding))) 
	     (cond ((primitive? value)
		    ((primitive-compilator value) exp cenv))
		   ((variable? value)
		    (expand-unknown-call exp proc args goto? cenv))
		   ((generated? value)
		    (let ((var ((structure-ref environments environment-ref)
				(generated-env value)
				(generated-symbol value))))
		      (expand-unknown-call var proc args goto? cenv)))
		   (else
		    (error "unknown type of binding value ~S" value))))))))

(define lambda-node? (node-predicate 'lambda))

(define operator/begin (get-operator 'begin syntax-type))

(define (expand-redex source proc args cenv id)
  (destructure (((#f formals . body) (node-form proc)))
    (let ((vars (map (lambda (name)
		       (make-variable (name->symbol name) type/unknown))
		     formals)))
      (make-let-exp vars
		    (map (lambda (arg id)
			   (expand arg cenv (name->symbol id)))
			 args
			 formals)
		    (expand (make-node operator/begin `(begin . ,body))
			    (bind formals vars cenv)
			    id)
		    source))))

(define (expand-unknown-call source proc args goto? cenv)
  (make-call-exp (get-primop primop/unknown-call)
		 0
		 type/unknown
		 `(,(really-expand-node proc cenv #f)
		   ,(make-quote-exp (if goto?
					`(goto ,(length args))
					(length args))
				    #f)
		   . ,(map (lambda (arg) (expand arg cenv #f))
			   args))
		 source))

;(define primitive-expanders (make-table))
;
;(define-scheme-syntax call-primitively exp (cenv id)
;  (destructure (((#f primitive . args) exp))
;    (let ((args (expand-list args cenv))
;	  (source (pair-annotation exp)))
;      (cond ((table-ref primitive-expanders primitive)
;	     => (lambda (proc)
;		  (proc source args cenv)))
;	    (else
;	     (let ((primop (get-scheme-primop primitive)))
;	       (make-call-exp primop 0 type/unknown args source)))))))

;(define (define-primitive-expander name nargs proc)
;  (table-set! primitive-expanders
;	      name
;	      (if nargs
;		  (lambda (source args cenv)
;		    (if (= nargs (length args))
;			(proc source args cenv)
;			(error "wrong number of arguments to primitive ~S" name)))
;		  proc)))

(define (expand-list exps cenv)
  (map (lambda (exp)
	 (expand exp cenv #f))
       exps))

; LETREC

; 1. Parse the specs
; 2. Expand the body (this must happen before the variable subrecords are added)
; 3. Build a subrecord for each variable
; 4. Expand the values keeping track of the reference graph in the subrecords
; 5. Sort the variable+value reference graph into strongly connected components
; 6. Build the actual expression
; 7. Flush the subrecords

; (letrec 1 (lambda (-vars-)
;             (set-contents! var val)
;             ...
;             body))
;
; (letrec -vars- -vals- body)

(define-scheme-syntax (letrec specs . body) (cenv id)
  (expand-letrec specs body cenv id))

(define (expand-letrec specs body cenv id)
  (receive (formals vars vals)
      (parse-labels-specs specs)
    (let* ((cenv (bind formals vars cenv))
	   (body (expand-block body cenv id)))
      (for-each (lambda (v)		; specs is used as a marker
		  (set-variable-flag! v (make-letrec-data specs)))
		vars)
      (for-each (lambda (var value)
		  (expand-letrec-value var value cenv vars specs))
		vars
		vals)
      (let ((sets (strongly-connected-components vars
						 variable-uses
						 variable-seen?
						 set-variable-seen?!)))
	(do ((sets sets (cdr sets))
	     (exp body (build-letrec (car sets) exp)))
	    ((null? sets)
	     (for-each (lambda (v) (set-variable-flag! v #f))
		       vars)
	     exp))))))

; Note that the formals are NOT unpainted.  This is important.

(define (parse-labels-specs specs)
  (let loop ((specs specs)
	     (formals '()) (vars '()) (vals '()))
    (if (null? specs)
	(values formals vars vals)
	(destructure ((((id value) . specs) specs))
	  (let ((var (make-variable (desyntaxify id) type/unknown)))
	    (loop specs
		  (cons id formals)
		  (cons var vars)
		  (cons value vals)))))))

;(define operator/lambda (get-operator 'lambda syntax-type))

;(define (parse-labels-spec spec)
;  (if (pair? (car spec))
;      (values (caar spec)
;	      (make-node operator/lambda
;			 `(lambda ,(cdar spec) . ,(cdr spec))))
;      (values (car spec) (cadr spec))))

; If there is only one variable and its value doesn't reference it, then
; use a LET instead of a LETREC.

(define (build-letrec vars body)
  (cond ((and (null? (cdr vars))
	      (not (memq? (car vars) (variable-uses (car vars)))))
	 (make-let-exp vars (map variable-value vars) body #f))
	((every? (lambda (v)
		   (lambda-exp? (variable-value v)))
		 vars)
	 (make-letrec-exp vars (map variable-value vars) body #f))
	(else
	 (error "LETREC values are not LAMBDAS"))))

;	 (make-let-exp vars
;		       (append (map (lambda (var)
;				      (make-letrec-set var
;						       (variable-value var)
;						       #f))
;				    vars)
;			       body)
;		       (map (lambda (var) (make-letrec-cell)) vars))

; List of LETREC bound variables currently in scope.

(define *letrec-variables* '())

(define (note-letrec-var-use var)
  (cond ((not (variable-seen? var))
	 (set-variable-seen?! var #t)
	 (set! *letrec-variables* (cons var *letrec-variables*)))))

(define-subrecord variable letrec-data variable-flag
  (marker    ; a unique marker for this LETREC
   )
  (value     ; the expanded value of this variable
   uses      ; a list of variables that VALUE uses
   seen?     ; #T if this has been seen before during the current expansion
   ))

(define make-letrec-data letrec-data-maker)

; Expand VALUE and determine which of VARS it uses.

(define (expand-letrec-value var value cenv vars mark)
  (let ((old-letrec-vars *letrec-variables*))
    (set! *letrec-variables* '())
    (for-each (lambda (v) (set-variable-seen?! v #f)) vars)
    (set-variable-value! var (expand value cenv (variable-name var)))
    (receive (ours others)
	(partition-list (lambda (var)
			  (eq? (variable-marker var) mark))
			*letrec-variables*)
      (set! *letrec-variables* (append others old-letrec-vars))
      (set-variable-uses! var ours))))


(define name-node? (node-predicate 'name 'leaf))

; Find lookup result that was cached by classifier

(define (name-node-binding node)
  (or (node-ref node 'binding)
      (node-form node)))  ; = (lookup cenv (node-form node))

; Ignore the value, it is used when running Pre-Scheme programs in Scheme

(define-scheme-syntax (external type value) (cenv id)
  (make-quote-exp (make-external-value (expand-type-spec type))
		  type/unknown))

(define (expand-type-spec spec)
  (cond ((pair? spec)
	 (case (car spec)
	   ((=>)
	    (make-arrow-type (map expand-type-spec (cadr spec))
			     (expand-type-spec (caddr spec))))
	   (else
	    (error "unknown type syntax ~S" spec))))
	((not (symbol? spec))
	 (error "unknown type syntax ~S" spec))
	((lookup-type spec)
	 => identity)
	(else
	 (error "unknown type name ~S" spec))))

