
; 1) Find all of the uvars that have no type
; 2) Find the ones that don't get values from anywhere, and the
;    ones that already have some value.
; 3) The ones that have values pass them on to others, and the others pass
;    them on, and so on.
; 4) The ones that don't get values from anywhere (for example arguments
;    to exported procedures that are not called here) get the smallest
;    type required by any of their uses.
; 5) Any necessary coercions are added.

; No source = no-one wants the value -- could happen

(define (add-type-coercions)
  (let* ((uvars (filter (lambda (uvar)
			  (not (uvar-binding uvar)))
			(constrained-uvars)))
	 (free (filter (lambda (uvar) (null? (uvar-subs uvar))) uvars))
	 (sources (filter constrained-source? uvars))) ; has side-effects
    (for-each flow-up-from-uvar sources)
    (for-each flow-down-to-uvar free)
    (for-each add-coercions uvars)))

(define (add-coercions uvar)
  (let ((type (uvar-binding uvar)))
    (for-each (lambda (relation)
		(let ((wants-type (maybe-follow-uvar (relation-dest relation))))
		  (if (not (eq? type wants-type))
		      (insert-coercion (relation-site relation)
				       type
				       wants-type))))
	      (uvar-supers uvar))
    (for-each (lambda (relation)
		(let ((source (relation-source relation)))
		  (if (and (not (uvar? source))
			   (not (eq? type source)))
		      (insert-coercion (relation-site relation)
				       source
				       type))))
	      (uvar-subs uvar))))

(define (insert-coercion site from to)
  (receive (ref set!)
      (cond ((pair?    site)    (values car set-car!))
	    ((form?    site)    (values form-value set-form-value!))
	    ((closure? site)    (values closure-exp set-closure-exp!))
	    ((lambda-exp? site) (values lambda-exp-body set-lambda-exp-body!))
	    (else
	     (error "coercion has a funny site ~S" site)))
    (let ((exp (ref site)))
      ; (format #t " coercing ~S from ~S to ~S~%" exp from to)
      (if (type> from to)
	  (begin (format #t "Warning: losing coercion ~S from ~S to ~S~%" exp from to)
		 ;(breakpoint "Losing coercion")
		 ))
      (cond ((quote-exp? exp)
	     (set-quote-exp-type! exp to))
	    (else
	     (let ((primop (get-scheme-primop 'coerce))
		   (args (list exp
			       (make-quote-exp from #f)
			       (make-quote-exp to   #f))))
	       (set! site (make-call-exp primop 0 to args #f))))))))

; Returns #T if none of UVAR's sources are other uvars or if it has a more
; than minimal initial type.  Also binds UVAR to its minimal type.

(define (constrained-source? uvar)
  (and (not (uvar-binding uvar))
       (not (null? (uvar-subs uvar)))
       (let loop ((rels (uvar-subs uvar))
		  (type (if (uvar-arith-loc uvar)
			    type/int32    ; smallest arithmetic type
			    #f))
		  (source? #t))
	 (cond ((null? rels)
		(if type (set-uvar-binding! uvar type))
		(or source? type))
	       (else
		(let ((source (relation-source (car rels))))
		  (cond ((uvar? source)
			 (loop (cdr rels) type #f))
			((type> source type)
			 (loop (cdr rels) source source?))
			(else
			 (loop (cdr rels) type source?)))))))))

; Find the smallest type that UVAR's customers want.

(define (flow-down-to-uvar uvar)
  (let loop ((rels (uvar-supers uvar)) (type #f))
    (cond ((null? rels)
	   (set-uvar-binding! uvar type)
	   type)
	  (else
	   (let ((type2 (flow-down-relation (car rels))))
	     (loop (cdr rels)
		   (if (or (not type)
			   (type> type type2))
		       type2
		       type)))))))

(define (flow-down-relation rel)
  (let ((dest (relation-dest rel)))
    (cond ((not (uvar? dest))
	   dest)
	  ((uvar-binding dest)
	   => identity)
	  (else
	   (flow-down-to-uvar dest)))))

; Pass UVAR's type to all of its customers.

(define (flow-up-from-uvar uvar)
  (let ((type (uvar-binding uvar)))
    ;(format #t "(flow-up-from-uvar ~S)~%" uvar)
    (for-each (lambda (rel)
		(depth-first-search rel type
				    relation-temp set-relation-temp!
				    flow-up-relation flow-up-loop))
	      (uvar-supers uvar))))

(define (flow-up-relation relation type do-next)
  (let ((dest (relation-dest relation)))
    (if (uvar? dest)
	(let ((has (uvar-binding dest)))
	  (cond ((or (not has)
		     (type> type has))
		 (set-uvar-binding! dest type)
		 (do-next (uvar-supers dest) type)))))))
  
(define (flow-up-loop relations type)
  (let ((has (uvar-binding (relation-dest (car relations)))))
    (if (type>= has type)
	(values #f #f)           ; continue from here
	(values #t type))))      ; restart from top of loop

; Depth-first search abstraction
; Loop is passed as a list of nodes in the order in which they were found.

(define-record-type stack
  ((nodes)
   data
   rest
   )
  ())

(define (depth-first-search start-node start-data
                            node-mark set-node-mark!
                            node-proc loop-proc)
  (let ((mark (cons 0 0)))
    (define (mark! node)
      (set-node-mark! node mark))
    (define (unmark! node)
      (set-node-mark! node #f))
    (define (do-node node data stack)
      (cond ((eq? mark (node-mark node))
	     (receive (loop remains)
		 (find-loop node stack)
	       (receive (restart? data)
		   (loop-proc loop data)
		 (cond (restart?
			(for-each unmark! loop)
			(do-node node data remains))
		       (else
			(next stack))))))
	    (else
	     (mark! node)
	     (node-proc node data (make-do-next stack))
	     (unmark! node)
	     (next stack))))
    (define (make-do-next stack)
      (lambda (nodes data)
	(if (null? nodes)
	    (next stack)
	    (do-node (car nodes) data (stack-maker nodes data stack)))))
    (define (next stack)
      (if (not stack)
	  (values)
	  (let ((nodes (cdr (stack-nodes stack))))
	    (cond ((null? nodes)
		   (next (stack-rest stack)))
		  (else
		   (set-stack-nodes! stack nodes)
		   (do-node (car nodes) (stack-data stack) stack))))))
    (do-node start-node start-data #f)))

(define (find-loop node stack)
  (let loop ((stack stack) (nodes '()))
    (if (eq? node (car (stack-nodes stack)))
	(values (cons node nodes) stack)
	(loop (stack-rest stack) (cons (car (stack-nodes stack)) nodes)))))

