; Find closures inside data structures and create new forms for them.

(define *forms* '())
(define *next-top-id* 0)

(define (flatten var value free-vars)
  (set! *forms* '())
  (set! *next-top-id* 0)
  (let ((new-var (cond ((stob? value) (stob-variable value))
		       ((closure? value) (closure-variable value))
		       (else #f))))
    (receive (value free)
	(if (or (not new-var)
		(eq? new-var var))
	    (values value (flatten-form-value! value))
	    (values new-var (list new-var)))
      (let ((forms *forms*))
	(set! *forms* '())
	(values value free forms)))))

(define (flatten-form-value! value)
  (cond ((stob? value)
	 (flatten-stob! value))
	((closure? value)
	 (or (closure-free value)
	     (flatten-closure! value)))
	((variable? value)
	 (list value))
	(else
	 '())))

(define (flatten-value value id)
  (cond ((not (or (closure? value)
		  (stob? value)))
	 value)
	((maybe-value->variable value)
	 => identity)
	(else
	 (let* ((var (generate-top-variable id))
		(form (make-form var value #f #f (flatten-form-value! value))))
	   (set! *forms* (cons form *forms*))
	   var))))

(define (maybe-value->variable value)
  (cond ((closure? value)
	 (closure-variable value))
	((stob? value)
	 (stob-variable value))
	(else
	 (bug "unknown value type ~S" value))))

(define (flatten-closure! closure)
  (for-each (lambda (p)
	      (let ((value (flatten-value (cdr p) (car p))))
		(set-variable-flag! (car p)
				    (if (variable? value)
					value
					(make-quote-exp value type/undetermined)))))
	    (closure-env closure))
  (let ((exp (closure-exp closure)))
    (receive (body vars free)
	(substitute-in-expression (lambda-exp-body exp)
				  (lambda-exp-vars exp))
      (for-each (lambda (p)
		  (set-variable-flag! (car p) #f))
		(closure-env closure))
      (set-closure-exp! closure
			(make-lambda-exp (lambda-exp-id exp)
					 (lambda-exp-return-type exp)
					 (lambda-exp-protocol exp)
					 vars
					 body
					 (lambda-exp-source exp)))

;      (if (let ((env (closure-env closure)))
;            (and (not (null? env))
;                 (variable? (caar env))
;                 (= 1751 (variable-id (caar env)))))
;          (breakpoint "flatten-closure"))
      free)))

(define (flatten-stob! stob)
  (let ((vector (stob-value stob)))
    (let loop ((i 0) (free '()))
      (if (>= i (vector-length vector))
	  free
	  (let ((val (flatten-value (vector-ref vector i) #f)))
	    (vector-set! vector i val)
	    (loop (+ i 1)
		  (if (or (not (variable? val))
			  (memq? val free))
		      free
		      (cons val free))))))))
		       
(define (generate-top-variable maybe-var)
  (let ((var (if maybe-var
		 (make-global-variable (variable-name maybe-var)
				       (variable-type maybe-var))
		 (make-global-variable (concatenate-symbol 'top. (next-top-id))
				       type/undetermined))))
    (set-variable-flags! var (cons 'generated-top-variable (variable-flags var)))
    var))

(define (next-top-id)
  (let ((id *next-top-id*))
    (set! *next-top-id* (+ 1 *next-top-id*))
    id))

(define (generated-top-variable? var)
  (memq? 'generated-top-variable (variable-flags var)))

;--------------------------------------------------

(define *free-exp-vars* #f)

(define (substitute-in-expression exp vars)
  (set! *free-exp-vars* '())
  (let* ((new-vars (copy-vars vars))
	 (exp (exp-substitute exp))
	 (free *free-exp-vars*))
    (set! *free-exp-vars* '())
    (set-variables-flag! new-vars #f)
    (set-variables-flag! free #f)
    (set-variables-flag! vars #f)
    (values exp new-vars free)))

(define (exp-substitute-list exps)
  (map exp-substitute exps))

(define (set-variables-flag! vars value)
  (for-each (lambda (v) (set-variable-flag! v value))
	    vars))

(define global-marker (list 'global))

(define (exp-substitute exp)
  (cond ((variable? exp)
	 (exp-substitute-variable exp))
	((quote-exp? exp)
	 exp)
	((call-exp? exp)
	 (make-call-exp (call-exp-proc exp)
			(call-exp-exits exp)
			(call-exp-type exp)
			(exp-substitute-list (call-exp-args exp))
			(call-exp-source exp)))
	((block-exp? exp)
	 (make-block-exp (exp-substitute-list (block-exp-exps exp))))
	((lambda-exp? exp)
	 (let* ((new-vars (copy-vars (lambda-exp-vars exp)))
		(body (exp-substitute (lambda-exp-body exp))))
	   (set-variables-flag! (lambda-exp-vars exp) #f)
	   (set-variables-flag! new-vars #f)
	   (make-lambda-exp (lambda-exp-id exp)
			    (lambda-exp-return-type exp)
			    (lambda-exp-protocol exp)
			    new-vars body
			    (lambda-exp-source exp))))
	((return-exp? exp)
	 (make-return-exp (return-exp-protocol exp)
			  (return-exp-type exp)
			  (exp-substitute-list (return-exp-args exp))))
	((let-exp? exp)
	 (let* ((vars (copy-vars (let-exp-vars exp)))
		(body (exp-substitute (let-exp-body exp)))
		(args (exp-substitute-list (let-exp-vals exp))))
	   (set-variables-flag! (let-exp-vars exp) #f)
	   (set-variables-flag! vars #f)
	   (make-let-exp vars args body (let-exp-source exp))))
	((letrec-exp? exp)
	 (let* ((vars (copy-vars (letrec-exp-vars exp)))
		(args (exp-substitute-list (letrec-exp-vals exp)))
		(body (exp-substitute (letrec-exp-body exp))))
	   (set-variables-flag! (letrec-exp-vars exp) #f)
	   (set-variables-flag! vars #f)
	   (make-letrec-exp vars args body (letrec-exp-source exp))))
	(else
	 (bug "unknown expression syntax ~S" exp))))

(define (exp-substitute-variable var)
  (let ((flag (variable-flag var)))
    (cond ((not flag)
	   (set! *free-exp-vars* (cons var *free-exp-vars*))
	   (set-variable-flag! var global-marker)
	   var)
	  ((eq? flag global-marker)
	   var)
	  (else
	   (exp-substitute flag)))))  ; because it too might be free

(define (copy-vars vars)
  (map (lambda (var)
	 (let ((new (copy-variable var)))
	   (set-variable-flag! var new)
	   (set-variable-flag! new global-marker)
	   new))
       vars))
