
(define-record-type closure
  (
   (exp)
   env
   )
  (
   (free #f)       ; variables free in BODY (top-level closures only)
   (variable #f)   ; global variable bound to this closure
   ))

(define make-closure closure-maker)

; Stored objects

(define-record-type stob
  (
   value
   )
  (
   type
   (variable #f)
   ))

(define make-stob stob-maker)

(define-record-type form
  (var              ; variable being defined (if any)
   (value)          ; current value
   source           ; one line of source code
   index            ; the index of this form in the source
   (free)           ; variables free in this form
   )

  (used?            ; is the value used in the program
   (exported? #f)   ; true if the definition in this form is exported
   (integrate 'okay) ; one of OKAY, NO, PARTIAL
   (aliases   '())  ; variables that are aliases for this one

   (dependency-index #f) ; index of this form in the data dependent order

   lambdas          ; list of all non-cont lambdas in this form

   (clients '())    ; forms that use this one's variable
   (providers '())  ; forms that define a variable used by this one
   
   (type #f)        ; one of LAMBDA, INTEGRATE, INITIALIZE or
                    ;  #F for unfinished forms
   merge            ; slot used by form-merging code
   temp             ; handy slot
   ))

(define-record-discloser type/form
  (lambda (form)
    `(form ,(form-index form)
	   ,(variable-name (form-var form)))))

(define (make-form var value source index free)
  (let ((source (if source (pp-one-line source) #f)))
    (let ((form (form-maker var value source index free)))
      (if (variable-form var)
	  (error "more than one definition of ~S" (variable-name var)))
      (set-variable-flags! var `((form . ,form) . ,(variable-flags var)))
      (cond ((closure? value)
	     (if (not (closure-variable value))
		 (set-closure-variable! value var)))
	    ((stob? value)
	     (if (not (stob-variable value))
		 (set-stob-variable! value var))))
      form)))

(define (pp-one-line x)
  (call-with-string-output-port
   (lambda (p)
     (write-one-line p 70 (lambda (p) (write x p))))))

(define (variable-form var)
  (cond ((flag-assq 'form (variable-flags var))
	 => cdr)
	(else #f)))

(define (set-form-node! form node lambdas)
  (set-form-value! form node)
  (set-form-lambdas! form lambdas)
  (if (node? node)
      (set-node-flag! node form)))

(define (erase-form form)
  (erase (form-value form))
  (set-form-node! form #f #f))

(define (make-form-current form)
  (initialize-lambdas)
  (add-lambdas (form-lambdas form))
  (set-node-flag! (form-value form) #f))

(define (form-name form)
  (if (form-var form)
      (variable-name (form-var form))
      (form-source form)))

(define (make-form-unused form)
  (set-form-type! form 'unused)
  (cond ((node? (form-value form))
	 (erase (form-value form))
	 (set-form-node! form #f '()))))

; notes on writing and reading forms
; What we really need here are forms.
; What to do?  Can read until there are no missing lambdas = end of form
; Need the variables as well.

; (form index type var source? clients providers integrate?)
; clients and providers are lists of indicies
; can get lambdas automatically

;(define (write-cps-file file forms)
;  (let ((port (make-tracking-output-port (open-output-file file))))
;    (reset-pp-cps)
;    (walk (lambda (f)
;	    (write-form f port))
;	  (sort-list forms
;		     (lambda (f1 f2)
;		       (< (form-index f1) (form-index f2)))))
;    (close-output-port port)))

;(define (write-form form port)
;  (format port "(FORM ~D ~S ~S "
;	  (form-index form)
;	  (form-type form)
;	  (form-integrate form))
;  (if (form-var form)
;      (print-variable-name (form-var form) port)
;      (format port "#f"))
;  (format port "~%  ~S" (map form-index (form-clients form)))
;  (rereadable-pp-cps (form-value form) port)
;  (format port ")~%~%"))

;------------------------------------------------------------------------------
; Put the forms that do not reference any other forms' variables in a queue.
; Every form gets a list of forms that use its variable and a list of forms
; whose variables it uses.

(define (sort-forms forms)
  (let ((queue (make-queue)))
    (for-each (lambda (f)
		(set-variable-flag! (form-var f) f))
	      forms)
    (let ((forms (really-remove-unreferenced-forms
		  forms
		  set-providers-using-free)))
      (for-each (lambda (f)
		  (if (null? (form-providers f))
		      (enqueue queue f)))
		(reverse forms))
      (for-each (lambda (f)
		  (set-variable-flag! (form-var f) #f))
		forms)
      (values forms (make-form-queue queue forms)))))

(define (set-providers-using-free form)
  (let loop ((vars (form-free form)) (provs '()))
    (cond ((null? vars)
	   (set-form-providers! form provs))
	  ((variable-flag (car vars))
	   => (lambda (prov)
		(set-form-clients! prov (cons form (form-clients prov)))
		(loop (cdr vars) (cons prov provs))))
	  (else
	   (loop (cdr vars) provs)))))

(define (make-form-queue ready forms)
  (let ((index 0))
    (lambda ()
      (let loop ()
	(cond ((not (queue-empty? ready))
	       (let ((form (dequeue ready)))
		 (set-form-dependency-index! form index)
		 (for-each (lambda (f)
			     (set-form-providers! f (delq! form (form-providers f)))
			     (if (and (null? (form-providers f))
				      (not (form-dependency-index f))
				      (form-used? f))
				 (enqueue ready f)))
			   (form-clients form))
		 (set! index (+ index 1))
		 form))
	      ((find-dependency-loop ready forms)
	       => (lambda (rest)
		    (set! forms rest)
		    (loop)))
	      (else #f))))))

; Find a circular dependence between the remaining forms.

(define (find-dependency-loop queue forms)
  (let ((forms (do ((forms forms (cdr forms)))
                   ((or (null? forms)
			(not (form-dependency-index (car forms))))
                    forms))))
    (cond ((null? forms)
           #f)
          (else
           ;;(format #t "Dependency loop!~%")
           (let ((form (really-find-dependency-loop forms)))
	     (if (not (every? (lambda (f) (eq? 'no (form-integrate f)))
			      (form-providers form)))
		 (set-form-integrate! form 'no))
             (set-form-providers! form '())
             (enqueue queue form)
             forms)))))

(define (really-find-dependency-loop forms)
  (for-each (lambda (f) (set-form-temp! f #f))
	    forms)
  (let label ((form (car forms)))
    (cond ((form-temp form)
           (break-dependency-loop (filter (lambda (f)
					    (and (form-temp f) (form-var f)))
					  forms)))
          (else
           (set-form-temp! form #t)
           (cond ((any-map label (form-providers form))
                  => (lambda (res)
                       (set-form-temp! form #f)
                       res))
                 (else
                  (set-form-temp! form #f)
                  #f))))))

(define (any-map proc list)
  (let loop ((list list))
    (cond ((null? list)
	   #f)
	  ((proc (car list))
	   => identity)
	  (else
	   (loop (cdr list))))))

(define *loop-forms* #f)

(define (break-dependency-loop forms)
  (or (first (lambda (f)
               (or (every? (lambda (f)
                             (eq? 'no (form-integrate f)))
                           (form-providers f))
                   (memq? f (form-providers f))))
             forms)
      (begin (set! *loop-forms* forms)
	     (let ((f (breakpoint "Break-dependency-loop ~S" forms)))
	       (set! *loop-forms* #f)
	       f))))

;------------------------------------------------------------------------------
; Turn expression into nodes and simplify it.

; Still to do:
;  Get representations of data values
;  Need to constant fold vector slots, including detection of modifications
;    and single uses.

(define (expand-and-simplify-form form)
  (let ((value (form-value form)))
    (cond ((closure? value)
	   (expand-and-simplify-lambda (closure-exp value) form))
	  ((variable-settable? (form-var form))  ; what if value is a lambda?
	   (set-form-type! form 'initialize)
	   "settable")
	  ((variable? value)
	   (add-known-form-value form value)
	   (cond ((variable-form value)
		  => (lambda (f)
		       (set-form-aliases! f
					  `(,(form-var form)
					    ,@(form-aliases form)
					    . ,(form-aliases f))))))
	   (set-form-type! form 'alias)
	   "alias")
	  ((stob? value)
	   (set-form-node! form (stob->node value) '())
	   (set-form-type! form 'stob)
	   "consed")
	  ((quote-exp? value)
	   (let ((value (quote-exp-value value)))
	     (cond ((external-value? value)
		    "external")
		   (else
		    (add-known-form-value form value)
		    "constant"))))
	  (else
	   (error "unknown kind of form value ~S ~S" value form)))))

; Make a call node containing the contents of the stob so that any
; variables will be seen as referenced and any integrable values will
; be integrated.

; Only works for vector stobs at this point.
; ALLOCATE is a randomly chosen primop, almost anything could be used.

(define (stob->node stob)
  (let* ((value (stob-value stob))
	 (contents '())
	 (add! (lambda (x) (set! contents (cons x contents)))))
    (cond ((vector? value)
	   (do ((i 0 (+ i 1)))
	       ((>= i (vector-length value)))
	     (add! (vector-ref value i))))
	  (else
	   (error "unknown kind of stob value ~S" value)))
    (let ((call (make-call-node (get-scheme-primop 'allocate)
				(+ 1 (length contents))
				0))
	  (node (make-lambda-node 'stob 'init '())))
      (attach call 0 (make-literal-node stob #f)) ; save for future use
      (do ((i 1 (+ i 1))
	   (cs (reverse contents) (cdr cs)))
	  ((null? cs))
	(let ((x (car cs)))
	  (attach call i (if (variable? x)
			     (make-reference-node x)
			     (make-literal-node x type/undetermined)))))
      (attach-body node call)
      (simplify-args call 1)
      node)))

(define (add-known-form-value form value)
  (let ((node (if (variable? value)
		  (make-reference-node value)
		  (make-literal-node value
				     (variable-type (form-var form))))))
    (add-known-value (form-var form) node)
    (erase node)
    (set-form-type! form 'integrate)))

(define (expand-and-simplify-lambda exp form)
  (receive (node lambdas)
      (expand-and-simplify-expression exp (form-name form))
;    (reconstruct-form-types form node)
    (set-form-value! form #f)
    (set-form-free! form #f)   ; old value no longer valid
    (let ((var (form-var form)))
;      (cond ((any? (lambda (n)
;                     (and (eq? 'proc (lambda-type n))
;                          (neq? node n)
;                          (not (no-free-references? n))))
;                   lambdas)
;             (if (neq? 'okay (form-integrate form))
;                 (error "~S must be integrated but some uses are already compiled"
;                        (form-name form)))
;             (format #t " (contains procedures)~%")
;;            (let ((name (symbol->string (variable-name var))))
;;              (if (and (< 6 (string-length name))
;;                       (string=? "proc##" (substring name 0 6)))
;;                  (breakpoint "(e-a-s-l ~A)" name)))
;             (add-known-value var node)
;             (erase node)
;             (set-form-type! form 'integrate))
;            (else  ))
      (let ((status (duplicate-form? form node)))
	(if status
	    (add-known-value var node))
	(set-form-node! form node lambdas)
	(set-form-type! form 'lambda)
	status))))

(define *duplicate-lambda-size* 10)

(define (duplicate-form? form node)
  (cond ((or (variable-settable? (form-var form))
	     (eq? 'no (form-integrate form)))
         #f)
        ((small-node? node *duplicate-lambda-size*)
	 "small")
        ((called-arguments? node)
         "called arguments")
        (else
         #f)))

(define (called-arguments? node)
  (any? (lambda (v)
	  (any? (lambda (n)
		  (eq? n (called-procedure-node (node-parent n))))
		(variable-refs v)))
	(cdr (lambda-variables node))))

;------------------------------------------------------------------------------

(define (integrate-stob-form form)
  (if (and (eq? 'stob (form-type form))
	   (elide-aliases! form)
	   (not (form-exported? form))
	   (every? cell-use (variable-refs (form-var form))))
      (let* ((var (form-var form))
	     (ref (car (variable-refs var)))
	     (call (lambda-body (form-value form))))
	; could fold any fixed references - do it later
	(cond ((and (null? (cdr (variable-refs var)))
		    (called-node? (cell-use ref)))
	       (format #t "computed-goto: ~S~%" (variable-name var)) 
	       (make-computed-goto form))))))

(define (cell-use node)
  (let ((parent (node-parent node)))
    (if (and (call-node? parent)
	     (eq? 'vector-ref (primop-id (call-primop parent))))
	parent
	#f)))

(define (primop-call? id node)
  (and (call-node? node)
       (eq? id (primop-id (call-primop node)))))

(define (elide-aliases! form)
  (not (or-map (lambda (f)
		 (switch-references! (form-var f) (form-var form))
		 (form-exported? f))
	       (form-aliases form))))

(define (switch-references! from to)
  (for-each (lambda (r)
	      (set-reference-variable! r to))
	    (variable-refs from))
  (set-variable-refs! to (append (variable-refs from) (variable-refs to))))

;------------------------------------------------------------------------------

(define (resimplify-form form)
  (let ((node (form-value form)))
    (cond ((and (node? node)
		(not (eq? 'stob (form-type form)))
		(not (node-simplified? node)))
	   (make-form-current form)
	   (simplify-node node)
	   (set-form-node! form node (make-lambda-list))))))

;------------------------------------------------------------------------------
; This is removes all forms that are not ultimately referenced from some
; exported form.

(define (add-form-provider! form provider)
  (if (not (memq? provider (form-providers form)))
      (set-form-providers!
       form
       (cons provider (form-providers form)))))

(define (variable->form var)
  (cond ((flag-assq 'form (variable-flags var))
	 => cdr)
	(else
	 (bug "variable ~S has no form" var))))

(define (maybe-variable->form var)
  (cond ((flag-assq 'form (variable-flags var))
	 => cdr)
	(else
	 #f)))

(define (remove-unreferenced-forms forms)
  (really-remove-unreferenced-forms forms set-form-providers))

(define (really-remove-unreferenced-forms forms set-form-providers)
  (receive (exported others)
      (partition-list form-exported? forms)
    (for-each (lambda (f)
		(set-form-providers! f '())
		(set-form-clients!   f '())
		(set-form-used?!     f (form-exported? f)))
	      forms)
    (for-each set-form-providers forms)
    (propogate-used?! exported)
    (append (remove-unused-forms others) exported)))

(define (set-form-providers form)
  (for-each (lambda (n)
	      (add-form-provider! (node-flag (node-base n)) form))
	    (variable-refs (form-var form)))
  (if (eq? (form-type form) 'alias)
      (add-form-provider! form (variable->form (form-value form)))))

(define (propogate-used?! forms)
  (let loop ((to-do forms))
    (if (not (null? to-do))
	(let loop2 ((providers (form-providers (car to-do)))
		    (to-do (cdr to-do)))
	  (if (null? providers)
	      (loop to-do)
	      (loop2 (cdr providers)
		     (let ((p (car providers)))
		       (cond ((form-used? p)
			      to-do)
			     (else
			      (set-form-used?! p #t)
			      (cons p to-do))))))))))

; Actually remove forms that are not referenced.

(define (remove-unused-forms forms)
  (format #t "Removing unused forms~%")
  (filter (lambda (f)
	    (cond ((or (not (form-used? f))
		       (let ((value (form-value f)))
			 (and (quote-exp? value)
			      (external-value? (quote-exp-value value)))))
		   (format #t " ~S~%" (variable-name (form-var f)))
		   (erase-variable (form-var f))
		   (cond ((node? (form-value f))
			  (erase (form-value f))
			  (set-form-node! f #f '())))
		   #f)
		  (else #t)))
	  forms))

;------------------------------------------------------------
; Total yucko.

; (unknown-call (lambda e-vars e-body)
;               protocol
;               (vector-ref x offset)
;               . args)
; =>
; (let (lambda ,vars
;        (computed-goto ,offset
;          ...
;          (lambda ()
;            (unknown-call (lambda ,copied-evars
;                            (jump ,(car vars) ,copied-evars))
;                          ,(vector-ref proc-vector i)
;                          . ,(cdr vars)))
;          ...))
;      ,exit
;      . ,args)

(define (make-computed-goto form)
  (let* ((ref (car (variable-refs (form-var form))))
	 (in-form (node-flag (node-base ref))))
    (make-form-current in-form)
    (really-make-computed-goto ref (call-args (lambda-body (form-value form))))
    (erase (form-value form))
    (set-form-node! form #f #f)
    (simplify-node (form-value in-form))
    (set-form-lambdas! in-form (make-lambda-list))))

(define (really-make-computed-goto use entries)
  (let* ((exits (- (vector-length entries) 1))
	 (vec-ref (node-parent use))
	 (offset (call-arg vec-ref 1))
	 (vector-call (node-parent vec-ref))
	 (args (sub-vector->list (call-args vector-call) 3))
	 (call (make-call-node (get-scheme-primop 'computed-goto)
			       (+ 1 exits)
			       exits))
	 (arg-vars (map (lambda (arg) (make-variable 't (node-type arg)))
			args))
	 (protocol (literal-value (call-arg vector-call 2)))
	 (cont (call-arg vector-call 0))
	 (cont-vars (lambda-variables cont))
	 (cont-type (make-arrow-type (map variable-type (lambda-variables cont))
				     type/null)))
    (detach cont)
    (for-each detach args)
    (let-nodes ((top (let 1 l1 cont . args))
		(l1 ((j cont-type) . arg-vars) call))
      (attach call exits (detach offset))
      (do ((i 0 (+ i 1)))
	  ((>= i exits))
	(attach call i (computed-goto-exit (detach (vector-ref entries (+ i 1)))
					   protocol arg-vars j cont-vars)))
      (replace-body vector-call top))))

(define (computed-goto-exit node protocol arg-vars cont-var cont-vars)
  (let* ((cont-vars (map copy-variable cont-vars))
	 (cont-args (map make-reference-node cont-vars))
	 (args (map make-reference-node arg-vars)))
    (let-nodes ((l1 () (unknown-call 1 l2 node '(protocol #f) . args))
		(l2 cont-vars (jump 0 (* cont-var) . cont-args)))
      l1)))
