
; Type Inference
;
; The entry points to the inferencer are:
;
; (unify! type1 type2 context)
;   Unify TYPE1 and TYPE2.  CONTEXT is used to provide user feedback when type
;   errors are detected.
;
; (make-uvar prefix depth . maybe-id)
;   Makes a new type variable.  PREFIX is a symbol, DEPTH is the current type
;   depth (used for polymorphism), and MAYBE-ID is an optional unique
;   integer.
;
; (make-constrained-type prefix type depth source)
;   Create a new type variable that is greater than or equal to TYPE.
;
; (make-constrained-arith-monop-type prefix arg-type1 depth source)
; (make-constrained-arith-binop-type prefix arg-type1 arg-type2 depth source)
;   Create new types associated with arithmetic operations.
;
; (schemify-type type depth)
;   Make TYPE polymorphic in any variables bound at DEPTH.
;
; (instantiate-type-scheme scheme depth)
;   Return an instantiation of SCHEME at DEPTH.
;
; (reset-inference!)
;   Clear various global variables (to be replaced with fluids at some point)


(define (unify! type1 type2 context)
  (cond ((really-unify! type1 type2)
	 => (lambda (status)
	      (unify-lost status type1 type2 context)))))

(define *currently-checking* #f)
(define *current-top-exp* #f)

(define atom-schemify (structure-ref syntactic schemify))

(define (schemify exp)
  (if (pair? exp)
      (map schemify exp)
      (atom-schemify exp)))

(define (unify-lost status type1 type2 context)
  (cond ((eq? context 'simplifying)
	 (bug "unification error while instantiating an integrable procedure"))
	((eq? context 'make-monomorphic)
	 #f)
	(else
	 (user-type-error-message status type1 type2 context))))

(define (user-type-error-message status type1 type2 context)
  (case (car status)
    ((arithmetic)
     (destructure (((#f uvar conflict value) status))
       (format #t "Type error: not an arithmetic type~% ")
       (display-type value (current-output-port))))
    ((circular)
     (destructure (((#f uvar value) status))
       (format #t "Type error: unification would make a circular type~% ")
       (display-type uvar (current-output-port))
       (format #t " == ")
       (display-type value (current-output-port))))
    ((mismatch)
     (destructure (((#f t1 t2) status))
       (format #t "Type error: type mismatch~% ")
       (display-type t1 (current-output-port))
       (format #t "~% ")
       (display-type t2 (current-output-port))))
    (else
     (bug "unknown unification error status ~S" status)))
  (cond ((and (call-exp? context)
	      (call-exp-source context))
	 => (lambda (source)
	      (format #t "~%The problem occurred with the call~% ~S" (schemify source))))
	((and *current-top-exp*
	      (find-some-source *current-top-exp* context))
	 => (lambda (source)
	      (format #t "~%The problem occurred within~% ~S" (schemify source)))))
  (if *currently-checking*
      (format #t "~%while reconstructing the type of '~S'" *currently-checking*))
  (error "type problem"))

(define (arith-type-error site)
  (let ((source (find-some-source *current-top-exp* site)))
    (format #t "Type error: arithmetic operation got a non-arithmetic argument~%")
    (if source
	(format #t "in ~S~%" (schemify source)))
    (if *currently-checking*
	(format #t "while reconstructing the type of '~S'" *currently-checking*))
    (error "type problem")))

(define (really-unify! p1 p2)
  (let ((p1 (maybe-follow-uvar p1))    ; get the current value of P1
	(p2 (maybe-follow-uvar p2)))   ; get the current value of P2
    (cond ((or (eq? p1 p2)
	       (eq? p1 type/null)
	       (eq? p2 type/null))
	   #f)
	  ((uvar? p1)
	   (maybe-bind-uvar! p1 p2))
	  ((uvar? p2)
	   (maybe-bind-uvar! p2 p1))
	  ((or (record-type? p2)
	       (record-type? p1))  ; just because I haven't implemented them yet
	   (error "unfinished implementation - really-unify!"))
	  ((other-type? p1)
	   (if (and (other-type? p2)
		    (eq? (other-type-kind p1) (other-type-kind p2))
		    (= (length (other-type-subtypes p1))
		       (length (other-type-subtypes p2))))
	       (unify-lists! (other-type-subtypes p1)
			     (other-type-subtypes p2))
	       (unify-failure 'mismatch p1 p2)))
	  (else
	   (unify-failure 'mismatch p1 p2)))))

(define (unify-lists! l1 l2)
  (let loop ((l1 l1) (l2 l2))
    (cond ((null? l1)
	   #f)
	  ((really-unify! (car l1) (car l2))
	   => identity)
	  (else
	   (loop (cdr l1) (cdr l2))))))

(define (unify-failure . args)
  args)

; Substitute VALUE for UVAR, if this will not introduce a circularity.

(define (maybe-bind-uvar! uvar value)
  (cond ((not (free-for? uvar value))
	 (unify-failure 'circular uvar value))
	((bind-uvar! uvar value)
	 => (lambda (conflict)
	      (unify-failure 'arithmetic uvar conflict value)))
	(else
	 #f)))

; Check that UVAR does not occur in EXP.

(define (free-for? uvar exp)
  (let label ((exp exp))
    (cond ((base-type? exp)
	   #t)
	  ((uvar? exp)
	   (if (eq? uvar exp)
	       #f
	       (if (uvar-binding exp)
		   (label (uvar-binding exp))
		   #t)))
	  ((other-type? exp)
	   (every? label (other-type-subtypes exp)))
	  (else
	   (identity (bug "funny type ~S" exp))))))
  
(define (type-conflict message . stuff)
  (apply breakpoint message stuff))

; For debugging
(define (uvar-name uvar)
  (concatenate-symbol (uvar-prefix uvar) "." (uvar-id uvar)))

; This should really be somewhere else.

; Instantiate TYPE and replace the types in NODE with their corresponding
; value.  LOCATION is where NODE will be applied, and is used to get the actual
; types of the arguments.

(define (instantiate-type&value type node location)
  (let ((has (instantiate-type-scheme type
				      -1
				      (lambda () (fix-types node))))
	(wants (call->proc-type (node-parent location))))
    (unify! has wants 'simplifying)))
;	  (format #t "~%Reconstructing ")
;	  (pp-cps call)
;	  (format #t " has   ~S~% wants ~S~%"
;		  (instantiate has)
;		  (instantiate wants))
;	  (breakpoint "reconstructing ~S" call)
;	  (unify! has wants 'simplifying)

; This is used to replace all references in NODE to polymorphic type variables
; with the current value of the type variable.
; Youch!  Very inefficient - may make many copies of the same type.

(define (fix-types node)
  (let label ((node node))
    (case (node-variant node)
      ((lambda)
       (for-each fix-variable (lambda-variables node))
       (label (lambda-body node)))
      ((call)
       (walk-vector label (call-args node)))
      ((literal)
       (let ((value (literal-value node)))
	 (if (or (uvar? value)
		 (other-type? value))
	     (set-literal-value! node (copy-type value))))))))

(define (fix-variable var)
  (set-variable-type! var (copy-type (variable-type var))))

(define (call->proc-type call)
  (make-arrow-type (do ((i (- (vector-length (call-args call)) 1) (- i 1))
			(ts '() (cons (maybe-instantiate
				       (node-type (call-arg call i)))
				      ts)))
		       ((< i 3)
			ts))
		   (let ((ret-vars (lambda-variables (call-arg call 0))))
		     (if (null? ret-vars)
			 type/unit
			 (variable-type (car ret-vars))))))

(define (maybe-instantiate type)
  (if (type-scheme? type)
      (instantiate-type-scheme type -1)
      type))

(define (make-monomorphic! var)
  (let ((type (type-scheme-type (variable-type var))))
    (for-each (lambda (ref)
		(cond ((not (called-node? ref))
		       (error
	  "polymorphic procedure ~S used as value, cannot be made monomorphic"
	  (variable-name var)))
		      ((unify! type
			       (call->proc-type (node-parent ref))
			       'make-monomorphic!)
 (error "polymorphic procedure ~S called on different types, can't translate to C"
	      (variable-name var)))))
	      (variable-refs var))
    (set-variable-type! var type)))