
(define-record-type base-type
  (name
   kind      ; INTEGER, FLOAT, or #f
   size      ; number of bytes
   )
  ())

(define-record-discloser type/base-type
  (lambda (base-type)
    (list (base-type-name base-type))))

(define base-type-table (make-table))

(define (make-base-type name kind size)
  (let ((type (base-type-maker name kind size)))
    (table-set! base-type-table name type)
    type))

(define (lookup-type id)
  (cond ((table-ref base-type-table id)
	 => identity)
	(else #f)))

;  0 <= x <= 127  (intersection of int8 and int8u, for constants)
(define type/int7u        (make-base-type 'int7u   'int   1))
(define type/int8         (make-base-type 'int8    'int   1))
(define type/int8u        (make-base-type 'int8u   'int   1))
(define type/int32        (make-base-type 'int32   'int   4))
(define type/float32      (make-base-type 'float32 'float 4))
(define type/float64      (make-base-type 'float64 'float 8))

(define type/null         (make-base-type 'null     #f #f))  ; no value
(define type/unit         (make-base-type 'unit     #f #f))  ; single value
(define type/boolean      (make-base-type 'boolean  #f 1))
(define type/undetermined (make-base-type '?        #f #f))
(define type/port         (make-base-type 'port     #f 4))

(define type/unknown type/undetermined) ; an alias

(define pointer-size 4) ; bytes in a pointer

(define (coercable-type? x)
  (and (base-type? x)
       (or (eq? (base-type-kind x) 'int)
	   (eq? (base-type-kind x) 'float))))

(define lattice-type? coercable-type?)

(define (type>= type1 type2)
  (let ((type1 (maybe-follow-uvar type1))
	(type2 (maybe-follow-uvar type2)))
    (or (eq? type1 type2)
	(if (base-type? type1)
	    (and (base-type? type2)
		 (base-type> type1 type2))
	    (type-eq? type1 type2)))))

(define (type> type1 type2)
  (not (type>= type2 type1)))

; This should probably be encoded in some more reasonable way.

(define (base-type> type1 type2)
  (if (or (not (base-type? type1))
	  (not (base-type? type2)))
      #f
      (case (base-type-name type1)
	((float32)
	 (not (eq? (base-type-name type2) 'float32)))
	((float16)
	 (eq? (base-type-kind type2) 'int))
	((int32)
	 (and (eq? (base-type-kind type2) 'int)
	      (> 4 (base-type-size type2))))
        ((int8 int8u)
	 (eq? (base-type-name type2) 'int7u))
	(else
	 #f))))

(define (base-type>= type1 type2)
  (not (base-type> type2 type1)))

(define (type-size type)
  (cond ((base-type? type)
	 (base-type-size type))
	((or (pointer-type? type)
	     (arrow-type? type))
	 pointer-size)
	(else
	 (error "type has no size ~S" type))))

(define (type-name type)
  (if (base-type? type)
      (base-type-name type)
      (error "type has no name ~S" type)))

; Return the join of two types (which must be EQ? or coercable types).

(define (join-types type1 type2)
  (cond ((or (not type2)
	     (eq? type1 type2))
	 type1)
	((or (not type1)
	     (base-type> type1 type2))
	 type2)
	(else
	 type1)))

;--------------------------------------------------
; This won't terminate on recursive types.

(define (type-eq? type1 type2)
  (let ((type1 (maybe-follow-uvar type1))
	(type2 (maybe-follow-uvar type2)))
    (or (eq? type1 type2)
	(and (other-type? type1)
	     (other-type? type2)
	     (eq? (other-type-kind type1)
		  (other-type-kind type2))
	     (let loop ((l1 (other-type-subtypes type1))
			(l2 (other-type-subtypes type2)))
	       (cond ((null? l1) (null? l2))
		     ((null? l2) #f)
		     ((type-eq? (car l1) (car l2))
		      (loop (cdr l1) (cdr l2)))
		     (else #f)))))))

;--------------------------------------------------
; Arrow, union, tagged union, and pointer types
; All done together to simplify the type walking

(define-record-type other-type
  (
   kind
   (subtypes)       ; set when finalized
   )
  (
   (finalized? #f)
   ))

(define make-other-type other-type-maker)

(define-record-discloser type/other-type
  (lambda (type)
    (case (other-type-kind type)
      ((arrow)
       (list 'arrow-type
	     (arrow-type-args type)
	     (arrow-type-result type)))
      (else
       (cons (other-type-kind type)
	     (other-type-subtypes type))))))

(define (make-other-type-predicate kind)
  (lambda (x)
    (and (other-type? x)
	 (eq? kind (other-type-kind x)))))

; Arrow

(define (make-arrow-type args result)
  (other-type-maker 'arrow (cons result args)))

(define arrow-type? (make-other-type-predicate 'arrow))

(define (arrow-type-result type)
  (car (other-type-subtypes type)))

(define (arrow-type-args type)
  (cdr (other-type-subtypes type)))

; Tuple

(define (make-tuple-type types)
  (other-type-maker 'tuple types))

(define tuple-type? (make-other-type-predicate 'tuple))
(define tuple-type-types other-type-subtypes)

; Pointer

(define (make-pointer-type type)
  (other-type-maker 'pointer (list type)))

(define pointer-type? (make-other-type-predicate 'pointer))
(define (pointer-type-to pointer-type)
  (car (other-type-subtypes pointer-type)))

(define type/char-pointer (make-pointer-type type/int8u))
(define type/byte-pointer (make-pointer-type type/int8u))

; Records

(define (make-record-type types)
  (other-type-maker 'record types))

(define record-type? (make-other-type-predicate 'record))
(define record-type-types other-type-subtypes)

;--------------------------------------------------

(define (finalize-type type)
  (let ((type (maybe-follow-uvar type)))
    (if (and (other-type? type)
	     (not (other-type-finalized? type)))
	(let ((subs (other-type-subtypes type)))
	  (set-other-type-finalized?! type #t)
	  (set-other-type-subtypes! type (map finalize-type subs))))
    type))

;--------------------------------------------------

;(define (erase-type type)
;  (cond ((uvar? type)
;	 (erase-uvar type))
;	((other-type? type)
;	 (for-each erase-type (other-type-subtypes type)))))


(define (print-relation-list list select)
  (for-each (lambda (rel)
	      (let ((dest (maybe-follow-uvar (select rel))))
		(if (uvar? dest)
		    (format #t " ~S~S" (uvar-prefix dest) (uvar-id dest))
		    (format #t " ~S" dest))))
	    list)
  (newline))

(define (show-constraints)
  (for-each (lambda (uvar)
	      (cond ((not (uvar-binding uvar))
		     (format #t "~S~S~%" (uvar-prefix uvar) (uvar-id uvar))
		     (print-relation-list (uvar-supers uvar) relation-dest)
		     (print-relation-list (uvar-subs uvar) relation-source))))
	    *constrained-uvars*))
		     
(define (display-type type port)

  (define (do-list list)
    (write-char #\( port)
    (cond ((not (null? list))
	   (do-type (car list))
	   (for-each (lambda (type)
		       (write-char #\space port)
		       (do-type type))
		     (cdr list))))
    (write-char #\) port))

  (define (do-type type)
    (cond ((base-type? type)
	   (display (base-type-name type) port))
	  ((arrow-type? type)
	   (do-list (arrow-type-args type))
	   (display " -> " port)
	   (do-type (arrow-type-result type)))
	  ((pointer-type? type)
	   (let ((to (pointer-type-to type)))
	     (write-char #\* port)
	     (cond ((arrow-type? to)
		    (write-char #\( port)
		    (do-type to)
		    (write-char #\) port))
		   (else
		    (do-type to)))))
	  ((uvar? type)
	   (write-char #\T port)
	   (display (uvar-id type) port))
	  ((type-scheme? type)
	   (display "(for-all " port)
	   (do-list (type-scheme-free-uvars type))
	   (display " " port)
	   (do-type (finalize-type (type-scheme-type type)))
	   (display ")" port))
	  (else
	   (bug "don't know how to display type ~S" type))))

  (do-type (finalize-type type)))