
; Call JOIN-SUBSTITUTE on all variable/value pairs.

(define (substitute-join-arguments lambda-proc call)
  (let ((vec (call-args call))
        (vars (lambda-variables lambda-proc)))
    (do ((vars vars (cdr vars))
         (i    1   (+ i 1))
         (c?   #f  (or (join-substitute (car vars) (vector-ref vec i))
                        c?)))
        ((null? vars) c?))))

; Does VAL take only one argument and is that argument passed to $TEST?
; Is VAR applied to constants?
; Are all references to the argument are below the $TEST call?
; Does the LCA of the references to VAR not contain any side-effects?
; Is VAR necessarily called in the LCA?
;
; If so, make the transformation described below.
;
; This is based on the assumption that the LCA of the references to VAR are
; small compared to VAL.  The opposite might be true, in which case the
; transformation will probably not be triggered.

(define (join-substitute var val)
  (let ((ref (and (lambda-node? val)
		  (simple-test-procedure val))))
    (and ref
	 (applied-to-useful-value? var ref)
	 (let ((lca (least-common-ancestor (variable-refs var))))
	   (cond ((or (suitable-join-conditional? lca var)
		      (suitable-join-point? val (node-parent ref)))
		  (really-join-substitute var val lca (node-parent ref))
		  #t)
		 (else #f))))))

; Check that VAL (a lambda-node) takes one argument, is jumped to, tests its
; argument, and that all references to the argument are at or below the test.

(define (simple-test-procedure val)
  (let ((vars (lambda-variables val)))
    (if (or (null? vars)
	    (not (null? (cdr vars)))
	    (not (car vars))
	    (not (calls-known? val))
	    (neq? 'jump (lambda-type val)))
	#f
	(let* ((var (car vars))
	       (ref (any simple-cond-ref (variable-refs var))))
	  (if (and ref (all-refs-below? var (node-parent (node-parent ref))))
	      ref
	      #f)))))

(define (simple-cond-ref ref)
  (if (primop-conditional? (call-primop (node-parent ref)))
      ref
      #f))

(define (all-refs-below? var node)
  (set-node-flag! node #t)
  (set-node-flag! (variable-binder var) #t)
  (let ((res (every? (lambda (r)
		       (eq? node (marked-ancestor r)))
		     (variable-refs var))))
    (set-node-flag! node #f)
    (set-node-flag! (variable-binder var) #f)
    res))

; Is VAR applied to something that can be used to simplify the conditional?

(define (applied-to-useful-value? var ref)
  (let ((call (node-parent ref))
	(index (node-index ref)))
    (any? (lambda (r)
	    (simplify-conditional? call index (call-arg (node-parent r) 1)))
	  (variable-refs var))))

; CALL is the least-common-ancestor of the references to VAR.  Check that
; the tree rooted at CALL contains no side-effects and that the control flow
; necessarily passes to VAR.  (Could check for undefined-effect here...)
; could do check that jumped-to proc if not VAR jumped to VAR eventually

(define (suitable-join-conditional? call var)
  (let label ((call call))
    (cond ((call-side-effects? call)
	   #f)
	  ((= 0 (call-exits call))
	   (and (eq? 'jump (primop-id (call-primop call)))
		(eq? var (reference-variable (called-procedure-node call)))))
	  (else
	   (let loop ((i 0))
	     (cond ((>= i (call-exits call))
		    #t)
		   ((not (label (lambda-body (call-arg call i))))
		    #f)
		   (else
		    (loop (+ i 1)))))))))

; #t if CALL performs side-effects.  The continuations to CALL are ignored.

(define (call-side-effects? call)
  (or (primop-side-effects (call-primop call))
      (let loop ((i (call-exits call)))
	(cond ((>= i (call-arg-count call))
	       #f)
	      ((side-effects? (call-arg call i))
	       #t)
	      (else
	       (loop (+ i 1)))))))

; The alternative to the above test: does the join point contain no side-effects
; above the test?

(define (suitable-join-point? join test)
  (let label ((call (lambda-body join)))
    (cond ((eq? call test)
	   #t)
	  ((call-side-effects? call)
	   #f)
	  (else
	   (let loop ((i 0))
	     (cond ((>= i (call-exits call))
		    #t)
		   ((not (label (lambda-body (call-arg call i))))
		    #f)
		   (else
		    (loop (+ i 1)))))))))
	     

; (let ((j (lambda (v)                  ; VAR VAL
;            .a.
;            ($test c1 c2 ... v ...)    ; TEST
;            .b.)))
;   .c.
;   (... (j x) ...)                     ; CALL
;   .d.)
; ==>
; .c.
; (.a.
;  (let ((v1 (lambda (x) c1[x/v]))
;        (v2 (lambda (x) c2[x/v])))
;    (... ($test (lambda () (v1 x)) (lambda () (v2 x)) ... x ...) ...))
;  .b.)
; .d.
;
; CALL is the least common ancestor of the references to VAR, which is bound to
; VAL, a procedure. TEST is a conditional that tests the argument passed to
; VAL.  This moves the body of VAL to the location of CALL and replaces
; TEST with CALL.  Calls to VAR are replaced with copies of TEST (which is
; the point of the whole exercise).

(define (really-join-substitute var val call test)
  (receive (cont-call cont-vars conts)
           (move-continuations test call)
    (let* ((value-var (car (lambda-variables val)))
	   (test-args (get-test-args test value-var))
	   (test-proc (call-primop test)))
      (parameterize-continuations conts value-var)
      (walk-refs-safely
       (lambda (ref)
	 (replace-join-var-with-test ref test-proc cont-vars conts test-args))
       var)
      (move-body cont-call
		 (lambda (cont-call)
		   (replace-body test cont-call)
		   (detach-body (lambda-body val))))
      (replace val (make-literal-node #f #f))
      (values))))

; Move the continuations of CALL to a LET call just above TO.  Returns a list
; of the variables now bound to the continuations and the continuations
; themselves.

(define (move-continuations call to)
  (let ((count (call-exits call)))
    (do ((i (- count 1) (- i 1))
	 (vs '() (cons (make-variable 'e (node-type (call-arg call i))) vs))
	 (es '() (cons (detach (call-arg call i)) es)))
	((< i 0)
	 (let ((new-call (make-call-node (get-primop primop/let) (+ count 1) 1))
	       (new-proc (make-lambda-node 'j 'cont vs)))
	   (attach-call-args new-call (cons new-proc es))
	   (insert-body new-call new-proc (node-parent to))
	   (values new-call vs es))))))

; Returns a list of the non-continuation arguments to TEST with #F in place of
; any that are references to VAR.

(define (get-test-args test var)
  (do ((i (- (call-arg-count test) 1) (- i 1))
       (as '() (cons (let ((node (call-arg test i)))
		       (cond ((and (reference-node? node)
				   (eq? var (reference-variable node)))
			      (erase (detach node))
			      #f)
			     (else
			      (detach node))))
		     as)))
      ((< i (call-exits test))
       as)))
		 
; Add a new variable to each of CONTS and substitute a reference to the correct
; variable for each reference to VAR within CONTS.

(define (parameterize-continuations conts var)
  (walk (lambda (n)
	  (let ((var (copy-variable var)))
	    (set-lambda-variables! n (cons var (lambda-variables n)))
	    (set-variable-binder! var n)
	    (set-node-flag! n #t)))
	conts)
  (walk-refs-safely
   (lambda (n)
     (let ((cont (marked-ancestor n)))
       (if cont
	   (replace n (make-reference-node (car (lambda-variables cont)))))))
   var)
  (walk (lambda (n) (set-node-flag! n #f)) conts)
  (values))

; REF is a reference to a variable that is being applied to one argument.
; Replace its parent with a call to TEST-PROC using copies of CONTS as
; continuations and TEST-ARGS as the additional arguments.  Any of TEST-ARGS
; that are #F are replaced with copies of the argument passed to REF.

(define (replace-join-var-with-test ref test-proc cont-vars conts test-args)
  (let* ((call (node-parent ref))
	 (arg (call-arg call 1))
         (new-call (make-call-node test-proc
				   (+ (length conts) (length test-args))
				   (length conts))))
    (attach-call-args new-call
		      (append (map (lambda (var cont)
				     (make-simple-jump-cont var cont arg))
				   cont-vars
				   conts)
			      (map (lambda (n)
				     (copy-node-tree (if n n arg)))
				   test-args)))
    (replace-body call new-call)))

; Return a lambda of no arguments whose body calls PROC (which is bound to VAR)
; on a copy of ARG.

(define (make-simple-jump-cont var proc arg)
  (let-nodes ((cont () call)
	      (call (jump 0 (* var) (! (copy-node-tree arg)))))
    cont))